
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.


"""

Twisted Pair: The framework of your ethernet.

Low-level networking transports and utilities.

See also twisted.protocols.ethernet, twisted.protocols.ip,
twisted.protocols.raw and twisted.protocols.rawudp.

Maintainer: Tommi Virtanen

"""

from twisted.pair._version import version
__version__ = version.short()
