# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

# 
'''
The Twisted Documentation Generation System

Maintainer: Andrew Bennetts
'''

# TODO
# Abstract
# Bibliography
# Index
# Allow non-web image formats (EPS, specifically)
# Allow pickle output and input to minimize parses
# Numbered headers
# Navigational aides

from twisted.lore._version import version
__version__ = version.short()
